@@ -1,8 +1,8 @@
 from ..generic import Tuple
+from ..json import to_json
 from ..scalar.ref import form_of, is_ref, Post, Put
 from ..state import State
-from ..uri import uri, URI
-from ..context import to_json
+from ..uri import URI
 
 
 class Column(object):
@@ -14,7 +14,7 @@ class Column(object):
         self.max_size = max_size
 
     def __eq__(self, other):
-        return self.name == other.name and self.dtype == other.dtype and self.max_size == other.max_size
+        return self.name == other.name and self.max_size == other.max_size and self.dtype is other.dtype
 
     def __json__(self):
         if self.max_size is None:
@@ -32,7 +32,7 @@ class Column(object):
 class Collection(State):
     """Data structure responsible for storing a collection of :class:`Value`s."""
 
-    __uri__ = uri(State) + "/collection"
+    __uri__ = URI(State) + "/collection"
 
     @classmethod
     def copy_from(cls, schema, source):
@@ -51,7 +51,7 @@ class Collection(State):
                 table = tc.table.Table.copy_from(table_schema, btree.keys())
         """
 
-        return cls(Post(uri(cls) + "/copy_from", {"schema": schema, "source": source}))
+        return cls(Post(URI(cls) + "/copy_from", {"schema": schema, "source": source}))
 
     @classmethod
     def load(cls, schema, data):
@@ -73,15 +73,12 @@ class Collection(State):
             raise ValueError(f"cannot load data {data} (consider calling `copy_from` instead)")
 
         class Load(cls):
-            def __init__(self, put_op):
-                cls.__init__(self, put_op)
-
             @property
             def schema(self):
                 return schema
 
             def __ref__(self, name):
-                return cls(URI(name))
+                return cls(name if isinstance(name, URI) else URI(name))
 
         return Load(Put(cls, schema, data))
 
