@@ -1,11 +1,11 @@
 """An n-dimensional array of numbers."""
 
 import inspect
-import logging
 import math
+import typing
 
 from ...decorators import post
-from ...generic import Map, Tuple
+from ...generic import autobox, gcs, resolve_class, Map
 from ...interface import Compare, Interface
 from ...math.interface import Boolean, Numeric, Trigonometric
 from ...math.operator import deref, is_one, is_zero, operator
@@ -18,11 +18,14 @@ from ...scalar.number import Bool, F32, F64, Number, U64
 from ...scalar import ref
 from ...shape import Shape
 from ...state import Class, State, Stream
-from ...uri import uri
+from ...uri import URI
 
 from ..base import Collection
 
-from .operator import Broadcast, Concatenate, Copy, Expand, Flip, Norm, Read, Reshape, Slice, Sum, Transpose
+from .operator import Broadcast, Cast, Concatenate, Copy, Expand, Flip, Norm, Read, Reshape, Slice, Sum, Transpose
+
+
+DType = typing.TypeVar("DType", bound=type[Number])
 
 
 class NDArray(Interface):
@@ -38,19 +41,19 @@ class NDArray(Interface):
 
     @property
     def dtype(self):
-        return Class(ref.Get(ref.MethodSubject(self, "dtype")))
+        return Class(ref.Get(URI(self, "dtype")))
 
     @property
     def ndim(self):
-        return U64(ref.Get(ref.MethodSubject(self, "ndim")))
+        return U64(ref.Get(URI(self, "ndim")))
 
     @property
     def shape(self):
-        return Shape(ref.Get(ref.MethodSubject(self, "shape")))
+        return Shape(ref.Get(URI(self, "shape")))
 
     @property
     def size(self):
-        return U64(ref.Get(ref.MethodSubject(self, "size")))
+        return U64(ref.Get(URI(self, "size")))
 
     def all(self):
         """Return `True` if all elements in this :class:`NDArray` are nonzero."""
@@ -69,12 +72,12 @@ class NDArray(Interface):
         If no `axis` is given, the total offset will be returned.
         """
 
-        return ref.Get(ref.MethodSubject(self, "argmax"), axis)
+        return ref.Get(URI(self, "argmax"), axis)
 
     def broadcast(self, shape):
         """Broadcast this :class:`NDArray` into the given `shape`."""
 
-        return ref.Get(ref.MethodSubject(self, "broadcast"), shape)
+        return ref.Get(URI(self, "broadcast"), shape)
 
     def cast(self, number_type):
         """Cast the data type of this :class:`NDArray` into the given `number_type`."""
@@ -84,12 +87,12 @@ class NDArray(Interface):
     def copy(self):
         """Return a copy of this :class:`NDArray`"""
 
-        return ref.Post(uri(Tensor) + "/copy_from", {"tensor": self})
+        return ref.Post(URI(Tensor) + "/copy_from", {"tensor": self})
 
     def expand_dims(self, axis=-1):
         """Expand the given `axis` of this :class:`NDArray`, or append a new axis if no `axis` is given."""
 
-        return ref.Get(ref.MethodSubject(self, "expand_dims"), axis)
+        return ref.Get(URI(self, "expand_dims"), axis)
 
     def flip(self, axis):
         """Flip this :class:`NDArray` along the given `axis`"""
@@ -101,14 +104,14 @@ class NDArray(Interface):
         Find the maxima of this :class:`NDArray` along the given `axis`, or the entire array if no `axis` is given.
         """
 
-        return ref.Get(ref.MethodSubject(self, "max"), axis)
+        return ref.Get(URI(self, "max"), axis)
 
     def min(self, axis=None):
         """
         Find the minima of this :class:`NDArray` along the given `axis`, or the entire array if no `axis` is given.
         """
 
-        return ref.Get(ref.MethodSubject(self, "min"), axis)
+        return ref.Get(URI(self, "min"), axis)
 
     def mean(self, axis=None):
         """
@@ -145,7 +148,7 @@ class NDArray(Interface):
         it will be the vector norm along that `axis`.
         """
 
-        return ref.Post(ref.MethodSubject(self, "norm"), _reduce_args(axis, keepdims))
+        return ref.Post(URI(self, "norm"), _reduce_args(axis, keepdims))
 
     def product(self, axis=None):
         """
@@ -153,17 +156,17 @@ class NDArray(Interface):
         or the total product if no `axis` is given.
         """
 
-        return ref.Get(ref.MethodSubject(self, "product"), axis)
+        return ref.Get(URI(self, "product"), axis)
 
     def reshape(self, shape):
         """Reshape this :class:`NDArray` into the given `shape`."""
 
-        return ref.Get(ref.MethodSubject(self, "reshape"), shape)
+        return ref.Get(URI(self, "reshape"), shape)
 
     def slice(self, bounds):
         """Return the sub-array of this :class:`NDArray` within the given `bounds`."""
 
-        return ref.Get(ref.MethodSubject(self), handle_bounds(bounds))
+        return ref.Get(URI(self), handle_bounds(bounds))
 
     def std(self, axis=None):
         """
@@ -181,7 +184,7 @@ class NDArray(Interface):
     def sum(self, axis=None, keepdims=False):
         """Compute the sum of this :class:`NDArray` along the given `axis`, or the total sum if no `axis` is given."""
 
-        return ref.Post(ref.MethodSubject(self, "sum"), _reduce_args(axis, keepdims))
+        return ref.Post(URI(self, "sum"), _reduce_args(axis, keepdims))
 
     def transpose(self, permutation=None):
         """
@@ -190,7 +193,7 @@ class NDArray(Interface):
         If no `permutation` is given, this will reverse the order of the axes of this :class:`NDArray`.
         """
 
-        return ref.Get(ref.MethodSubject(self, "transpose"), permutation)
+        return ref.Get(URI(self, "transpose"), permutation)
 
     def write(self, value):
         """Overwrite this :class:`NDArray` with the given :class:`NDArray` or `Number`, broadcasting if needed."""
@@ -198,11 +201,10 @@ class NDArray(Interface):
         return self._put("", None, value)
 
 
-class Tensor(Collection, NDArray, Trigonometric, Boolean, Numeric, Compare):
+class Tensor(Collection, NDArray, Trigonometric, Boolean, Numeric, Compare, typing.Generic[DType]):
     """An n-dimensional array of numbers."""
 
-    __uri__ = uri(Collection) + "/tensor"
-    __spec__ = (Shape, Number)
+    __uri__ = URI(Collection) + "/tensor"
 
     def __init__(self, form):
         if isinstance(form, Number) or isinstance(deref(form), (bool, float, int)):
@@ -210,101 +212,6 @@ class Tensor(Collection, NDArray, Trigonometric, Boolean, Numeric, Compare):
 
         Collection.__init__(self, form)
 
-    @classmethod
-    def trig_rtype(cls):
-        shape, dtype = cls.__spec__
-        return cls.expect(shape, dtype.trig_rtype())
-
-    @classmethod
-    def expect(cls, shape, dtype):
-        """
-        Define a new subclass of `cls` which captures the given shape and datatype.
-
-        It would be better to implement this feature using generic type parameters (i.e. `class Tensor[Shape, DType]:`)
-        but this is not supported on Python <= 3.8.
-        """
-
-        if inspect.isclass(shape):
-            if shape is not Tuple and shape is not Shape:
-                raise ValueError(f"invalid type for tensor shape: {shape}")
-
-        elif not isinstance(shape, (list, tuple, Tuple)):
-            raise ValueError(f"invalid tensor shape: {shape}")
-
-        spec = (shape, dtype)
-
-        if not isinstance(shape, Shape):
-            shape = Shape(shape)
-
-        class _Tensor(cls):
-            __spec__ = spec
-
-            def __init__(self, form):
-                if ref.is_literal(shape) and operator(form):
-                    try:
-                        actual_shape = operator(form).shape
-                        if len(shape) != len(actual_shape) or not all(e == a for e, a in zip(shape, actual_shape)):
-                            raise ValueError(f"wrong shape for {self}: {actual_shape} (expected {shape})")
-                    except (RuntimeError, ValueError) as e:
-                        logging.debug(lambda: f"{form} does not have a literal shape: {e}")
-
-                Tensor.__init__(self, form)
-
-            @classmethod
-            def create(cls):
-                op_ref = ref.Get(cls, (shape, dtype))
-                return cls(op_ref)
-
-            @property
-            def dtype(self):
-                return dtype
-
-            @property
-            def ndim(self):
-                if hasattr(shape, "__len__"):
-                    return len(shape)
-                else:
-                    return ref.Get(ref.MethodSubject(self, "ndim"))
-
-            @property
-            def schema(self):
-                return shape, dtype
-
-            @property
-            def shape(self):
-                return shape
-
-        return _Tensor
-
-    @classmethod
-    def create(cls, shape, dtype=F32):
-        """Create a new, empty `Tensor`. Call this method to initialize a persistent `Tensor` in a `Chain`."""
-
-        return cls.expect(shape, dtype).create()
-
-    @classmethod
-    def load(cls, shape, data, dtype=F32, name=None):
-        """
-        Load a `Tensor` from an existing data set.
-
-        The `name` parameter is useful in the string representation of an operator graph.
-
-        Example:
-            .. highlight:: python
-            .. code-block:: python
-
-                coords = [[0, 0, 1], [0, 1, 0]]
-                values = [1, 2]
-                sparse = tc.tensor.Sparse.load([2, 3, 4], zip(coords, values))
-                dense = tc.tensor.Dense.load([2, 3, 4], values, tc.I32)
-        """
-
-        name = name if name else f"load {shape}"
-
-        cls = cls.expect(shape, dtype)
-        op_ref = ref.Get(uri(cls) + "/load", ((shape, dtype), data), name)
-        return cls(op_ref)
-
     def __repr__(self):
         if operator(self):
             return repr(operator(self))
@@ -319,26 +226,70 @@ class Tensor(Collection, NDArray, Trigonometric, Boolean, Numeric, Compare):
 
         return Tensor(form=MatMul(self, other))
 
+    @classmethod
+    def trig_rtype(cls):
+        rtype = Number
+
+        if hasattr(cls, "__orig_class__"):
+            dtype = resolve_class(typing.get_args(cls.__orig_class__)[0])
+            rtype = dtype.trig_rtype() if inspect.isclass(dtype) and issubclass(dtype, Number) else Number
+
+        return Tensor[rtype]
+
+    @classmethod
+    def with_shape(cls, shape):
+        if not hasattr(shape, "__len__"):
+            return cls
+
+        class TensorWithExpectedShape(cls):
+            @property
+            def ndim(self):
+                return len(shape)
+
+            @property
+            def shape(self):
+                default = self._get("shape", rtype=Shape)
+                return Shape([shape[x] if ref.is_literal(shape[x]) else default[x] for x in range(len(shape))])
+
+        return TensorWithExpectedShape
+
+    @property
+    def dtype(self):
+        if hasattr(self, "__orig_class__"):
+            _shape, dtype = typing.get_args(self.__orig_class__)
+            dtype = resolve_class(dtype)
+            if inspect.isclass(dtype) and issubclass(dtype, Number):
+                return dtype
+
+        return Class[Number](form=ref.Get(URI(self, "dtype")))
+
     @property
     def ndim(self):
         """Return the number of dimensions of this `Tensor`."""
 
-        if hasattr(self.shape, "__len__"):
-            return len(self.shape)
+        shape = self.shape
+
+        if hasattr(shape, "__len__"):
+            return len(shape)
         else:
             return self._get("ndim", rtype=U64)
 
+    @property
+    def schema(self):
+        return self.shape, self.dtype
+
     @property
     def shape(self):
         """Return the shape of this `Tensor`."""
 
-        if operator(self):
-            try:
-                return operator(self).shape
-            except (RuntimeError, ValueError):
-                logging.debug(f"{self} does not have a literal shape")
+        default = self._get("shape", rtype=Shape)
 
-        return self._get("shape", rtype=Shape)
+        try:
+            return getattr(ref.deref(self), "shape", default)
+        except (RuntimeError, ValueError):
+            pass
+
+        return default
 
     def abs(self):
         return Tensor(form=Abs(self))
@@ -356,7 +307,10 @@ class Tensor(Collection, NDArray, Trigonometric, Boolean, Numeric, Compare):
         return Tensor(form=Broadcast(self, shape))
 
     def cast(self, dtype):
-        return Tensor(form=NDArray.cast(self, dtype))
+        try:
+            return Tensor[dtype](form=Cast(self, dtype))
+        except TypeError:
+            return Tensor(form=Cast(self, dtype))
 
     def copy(self):
         return Tensor(form=Copy(self))
@@ -487,7 +441,7 @@ class Tensor(Collection, NDArray, Trigonometric, Boolean, Numeric, Compare):
         if not ref.is_literal(copy):
             raise ValueError(f"reshape requires a literal boolean for copy, not {copy}")
 
-        reshaped = Tensor(form=Reshape(self, shape))
+        reshaped = Tensor(form=Reshape(self, Shape(shape)))
 
         if copy:
             return reshaped.copy()
@@ -517,6 +471,9 @@ class Tensor(Collection, NDArray, Trigonometric, Boolean, Numeric, Compare):
         return WritableView(Slice(self, bounds))
 
     def sub(self, other):
+        if ref.same_as(other, 0):
+            return self
+
         return Tensor(form=Sub(self, other))
 
     def sum(self, axis=None, keepdims=False):
@@ -535,7 +492,7 @@ class Tensor(Collection, NDArray, Trigonometric, Boolean, Numeric, Compare):
         return Tensor(form=Transpose(self, permutation))
 
 
-class Dense(Tensor):
+class Dense(Tensor, typing.Generic[DType]):
     """
     An n-dimensional array of numbers stored as sequential blocks.
 
@@ -544,86 +501,113 @@ class Dense(Tensor):
     you can safely treat this response as a divide-by-zero error.
     """
 
-    __uri__ = uri(Tensor) + "/dense"
+    __uri__ = URI(Tensor) + "/dense"
 
     @classmethod
-    def arange(cls, shape, start, stop, name=None):
+    def arange(cls, shape, start, stop):
         """
         Return a `Dense` tensor with the given shape containing a range of numbers evenly distributed
         between `start` and `stop`.
-
-        The `name` parameter is used only in the string representation of an operator graph.
         """
 
-        name = name if name else f"arange({start}, {stop})x{shape}"
-        dtype = type(start) if isinstance(start, Number) else Number
-        cls = cls.expect(shape, dtype)
-        op_ref = ref.Get(uri(cls) + "/range", (shape, start, stop), name)
-        return cls(op_ref)
+        return cls.with_shape(shape)(form=ref.Get(URI(cls, "range"), (shape, start, stop)))
 
     @classmethod
     def concatenate(cls, tensors, axis=0):
         """Create a new `Dense` tensor by concatenating the given `tensors` along the given `axis`."""
 
-        return Dense(form=Concatenate(tensors, axis))
+        return cls(form=Concatenate(tensors, axis))
 
     @classmethod
-    def constant(cls, shape, value, name=None):
+    def create(cls, shape, dtype=F32):
         """
-        Return a `Dense` tensor filled with the given `value`.
+        Create a new, empty :class:`Dense` tensor.
 
-        The `name` parameter is used only in the string representation of an operator graph.
+        Call this method to initialize a persistent `Tensor` in a `Chain`.
         """
 
-        name = name if name else f"{value}x{shape}"
-        dtype = type(value) if isinstance(value, Number) else Number
-        cls = cls.expect(shape, dtype)
-        op_ref = ref.Get(uri(cls) + "/constant", (shape, value), name)
-
-        if ref.same_as(value, 1):
-            return cls(op_ref)
-        if ref.same_as(value, 0):
-            return cls(op_ref)
-        else:
-            return cls(op_ref)
+        return cls[dtype].with_shape(shape)(form=ref.Get(URI(cls), (shape, dtype)))
 
     @classmethod
-    def ones(cls, shape):
-        """Construct a `Dense` tensor with dtype :class:`F64` filled with ones."""
+    def load(cls, shape, data, dtype=F32):
+        """
+        Load a :class:`Dense` tensor from an existing data set.
 
-        return cls.expect(shape, F64).constant(shape, 1.)
+        Example:
+            .. highlight:: python
+            .. code-block:: python
+
+                values = [0, 1, 2]
+                dense = tc.tensor.Dense.load([1, 3], values, tc.I32)
+        """
+
+        return cls[dtype].with_shape(shape)(form=ref.Get(URI(cls, "load"), ((shape, dtype), data)))
+
+    @classmethod
+    def constant(cls, shape, value):
+        """Return a `Dense` tensor of the given `shape` filled with the given `value`."""
+
+        assert not inspect.isclass(value)
+        value = autobox(value)
+        op_ref = ref.Get(URI(cls, "constant"), (shape, value))
+
+        if isinstance(value, Number):
+            try:
+                return cls[type(value)].with_shape(shape)(form=op_ref)
+            except TypeError:
+                pass
+
+        return cls.with_shape(shape)(form=op_ref)
+
+    @classmethod
+    def ones(cls, shape, dtype=F32):
+        """Construct a `Dense` tensor filled with ones."""
+
+        try:
+            return cls[dtype].constant(shape, dtype(1.))
+        except TypeError:
+            return cls.constant(shape, 1.)
 
     @classmethod
     def ones_like(cls, tensor):
-        """Return a `Dense` tensor filled with ones, with the same shape and data type as the given `tensor`."""
+        """Return a `Dense` tensor filled with ones, with the same shape as the given `tensor`."""
 
+        # TODO: include data type
         return cls.ones(tensor.shape)
 
     @classmethod
-    def zeros(cls, shape):
-        """Construct a `Dense` tensor with dtype :class:`F64` filled with ones."""
+    def zeros(cls, shape, dtype=F32):
+        """Construct a `Dense` tensor filled with zeros."""
 
-        return cls.expect(shape, F64).constant(shape, 0.)
+        try:
+            return cls[dtype].constant(shape, dtype(0.))
+        except TypeError:
+            return cls.constant(shape, 0)
 
     @classmethod
     def zeros_like(cls, tensor):
-        """Return a `Dense` tensor filled with zeros, with the same shape and data type as the given `tensor`."""
+        """Return a `Dense` tensor filled with zeros, with the same shape as the given `tensor`."""
 
+        # TODO: include data type
         return cls.zeros(tensor.shape)
 
     @classmethod
-    def random_normal(cls, shape, mean=0.0, std=1.0, name=None):
-        """
-        Return a `Dense` tensor filled with a random normal distribution of `F64` s.
+    def random_normal(cls, shape, mean=0.0, std=1.0):
+        """Return a `Dense` tensor filled with a random normal distribution of `F64` s."""
 
-        The `name` parameter is used only in the string representation of an operator graph.
-        """
+        class RandomNormal(cls[F64]):
+            @property
+            def shape(self):
+                default = self._get("shape", rtype=Shape)
+
+                if hasattr(shape, "__len__"):
+                    return Shape([shape[x] if ref.is_literal(shape[x]) else default[x] for x in range(len(shape))])
+                else:
+                    return default
 
-        name = name if name else f"random {shape}"
-        cls = cls.expect(shape, F64)
         args = {"shape": shape, "mean": mean, "std": std}
-        op_ref = ref.Post(uri(cls) + "/random/normal", args, name)
-        return cls(op_ref)
+        op_ref = ref.Post(URI(cls, "random/normal"), args)
+        return RandomNormal(form=op_ref)
 
     @classmethod
     def random_uniform(cls, shape, minval=0, maxval=1):
@@ -633,11 +617,12 @@ class Dense(Tensor):
             raise ValueError(f"Dense.random_uniform requires a literal range, not [{minval}, {maxval})")
 
         if minval == maxval:
-            return cls.constant(shape, minval)
+            return cls[F64].constant(shape, float(minval))
 
         assert maxval > minval
 
-        random = cls(ref.Get(uri(cls) + "/random/uniform", shape))
+        random = cls[F64].with_shape(shape)(form=ref.Get(URI(cls, "random", "uniform"), shape))
+
         if (minval, maxval) == (0, 1):
             return random
         else:
@@ -655,8 +640,11 @@ class Dense(Tensor):
         `minval` and `maxval` default to two standard deviations.
         """
 
+        if not ref.is_literal((mean, std)):
+            raise TypeError(f"truncated normal distribution requires literal parameters, not mean={mean} and std={std}")
+
         if not std:
-            return cls.constant(shape, mean)
+            return cls[F64].constant(shape, float(mean))
 
         minval = std * -2 if minval is None else minval
         maxval = std * 2 if maxval is None else maxval
@@ -675,7 +663,7 @@ class Dense(Tensor):
             return Map(tensor=cxt.new_tensor.copy())
 
         truncated = Map(ref.While(cond, step, Map(tensor=Dense.random_normal(shape, mean, std))))["tensor"]
-        return cls(truncated)
+        return cls[F64](form=truncated)
 
     def add(self, other):
         if ref.same_as(other, 0):
@@ -700,10 +688,13 @@ class Dense(Tensor):
         return self._get("elements", bounds, Stream)
 
     def sub(self, other):
+        if ref.same_as(other, 0):
+            return self
+
         return Dense(form=Tensor.sub(self, other))
 
 
-class Sparse(Tensor):
+class Sparse(Tensor, typing.Generic[DType]):
     """
     An n-dimensional array of numbers stored as a :class:`Table` of coordinates and values.
 
@@ -715,27 +706,64 @@ class Sparse(Tensor):
     which are not filled in the left-hand `Tensor`.
     """
 
-    __uri__ = uri(Tensor) + "/sparse"
+    __uri__ = URI(Tensor) + "/sparse"
 
     @classmethod
-    def zeros(cls, shape, dtype=F32, name=None):
+    def create(cls, shape, dtype=F32):
+        """
+        Create a new, empty :class:`Sparse` tensor.
+
+        Call this method to initialize a persistent :class:`Tensor` in a :class:`Chain`.
+        """
+
+        op_ref = ref.Get(URI(cls), (shape, dtype))
+
+        try:
+            return cls[dtype].with_shape(shape)(form=op_ref)
+        except TypeError:
+            return cls.with_shape(shape)(form=op_ref)
+
+    @classmethod
+    def load(cls, shape, data, dtype=F32):
+        """
+        Load a :class:`Sparse` tensor from an existing data set.
+
+        Example:
+            .. highlight:: python
+            .. code-block:: python
+
+                coords = [[0, 0, 1], [0, 1, 0]]
+                values = [1, 2]
+                sparse = tc.tensor.Sparse.load([2, 3, 4], zip(coords, values))
+        """
+
+        op_ref = ref.Get(URI(cls, "load"), ((shape, dtype), data))
+
+        try:
+            return cls[dtype].with_shape(shape)(form=op_ref)
+        except TypeError:
+            return cls.with_shape(shape)(form=op_ref)
+
+    @classmethod
+    def zeros(cls, shape, dtype=F32):
         """
         Return a `Sparse` tensor with the given shape and data type.
 
         If `dtype` is not specified, the data type will be :class:`F32`.
-
-        The `name` parameter is used only in the string representation of an operator graph.
         """
 
-        cls = cls.expect(shape, dtype)
-        name = name if name else f"sparse 0x{shape}"
-        return cls(ref.Get(cls, (shape, dtype), name))
+        op_ref = ref.Get(URI(cls), (shape, dtype))
+
+        try:
+            return cls[dtype].with_shape(shape)(form=op_ref)
+        except TypeError:
+            return cls.with_shape(shape)(form=op_ref)
 
     @classmethod
     def zeros_like(cls, tensor):
         """Return a `Sparse` tensor with the same shape and data type as the given `tensor`."""
 
-        return cls.zeros(tensor.shape)
+        return cls[tensor.dtype].zeros(tensor.shape, tensor.dtype)
 
     def as_dense(self):
         """Return a :class:`Dense` view of this `Sparse` tensor."""
