@@ -1,2 +1,2 @@
-from .app import Graph, Schema
+from .app import Graph, Schema, create_schema
 from .edge import DIM, Edge, ForeignKey, Vector
