@@ -1,9 +1,10 @@
-from tinychain.collection.tensor import einsum, Sparse
-from tinychain.decorators import closure, get_op, post_op, put_op, delete_op
-from tinychain.error import BadRequest
-from tinychain.ref import After, Get, If, MethodSubject, While, With
-from tinychain.state import Map, Tuple
-from tinychain.value import Bool, Nil, I32, U64, String
+from ..collection.tensor import einsum, Sparse
+from ..decorators import post
+from ..error import BadRequest
+from ..generic import Map
+from ..scalar.number import Bool, I32, U32, U64
+from ..scalar.ref import If, While
+from ..scalar.value import String
 
 
 ERR_DELETE = "cannot delete {{column}} {{id}} because it still has edges in the Graph"
@@ -38,41 +39,66 @@ class Schema(object):
 class Edge(Sparse):
     """A relationship between a primary key and itself."""
 
+    __spec__ = ((DIM, DIM), U32)
+
+    def link(self, from_id, to_id):
+        """Add a link from the :class:`Model` at `from_id` to the :class:`Model` at `to_id`"""
+
+        return self[from_id, to_id].write(True)
+
+    def unlink(self, from_id, to_id):
+        """Remove the link from the :class:`Model` at `from_id` to the :class:`Model` at `to_id`, if present."""
+
+        return self[from_id, to_id].write(False)
+
     def match(self, node_ids, degrees):
         """
-        Traverse this `Edge` breadth-first from the given `node_ids` for the given number of `degrees`.
+        Traverse this `Edge` breadth-first from the given `node_ids`.
 
-        Returns a new vector filled with the IDs of the matched nodes.
+        Returns a vector view with the IDs of the matched nodes.
+        Call `match` in a `While` loop to traverse multiple degrees of relationships.
         """
 
-        @post_op
+        @post
         def cond(i: U64):
             return i < degrees
 
-        @post_op
+        @post
         def traverse(edge: Sparse, i: U64, neighbors: Sparse):
-            neighbors += Sparse.sum(edge * neighbors, 1)
+            neighbors += (edge * neighbors).sum(1)
             return {"edge": edge, "i": i + 1, "neighbors": neighbors.copy()}
 
         node_ids = Sparse(node_ids)
         shape = node_ids.shape
         traversal = If(
-            shape.eq([DIM]),
+            node_ids.ndim == 1,
             While(cond, traverse, {"edge": self, "i": 0, "neighbors": node_ids}),
             BadRequest(String(f"an edge input vector has shape [{DIM}], not {{{shape}}}").render(shape=shape)))
 
-        return Sparse.sub(Map(traversal)["neighbors"], node_ids)
+        return Sparse(Map(traversal)["neighbors"]) - node_ids
 
 
 class ForeignKey(Sparse):
     """A relationship between a primary key and a column in another `Table`."""
 
-    def backward(self, node_ids):
+    __spec__ = ((DIM, DIM), U32)
+
+    def primary(self, node_ids):
         """Return a vector of primary node IDs, given a vector of foreign node IDs."""
 
         return einsum("ij,j->i", [self, node_ids])
 
-    def forward(self, node_ids):
+    def foreign(self, node_ids):
         """Return a vector of foreign node IDs, given a vector of primary node IDs."""
 
         return einsum("ij,i->j", [self, node_ids])
+
+
+class Vector(Sparse):
+    """A `Vector` of node IDs used to query an :class:`Edge` or `ForeignKey`"""
+
+    __spec__ = ((DIM,), Bool)
+
+    @classmethod
+    def create(cls):
+        return cls(cls.__spec__)
