@@ -1,8 +1,8 @@
 import logging
 import typing
 
-from ..context import deanonymize, to_json
-from ..scalar.ref import deref, hex_id, is_literal, same_as, is_op_ref, reference, Op
+from ..json import to_json
+from ..scalar.ref import deref, is_literal, same_as, is_op_ref, Op
 from ..scalar.value import Id
 
 from .base import is_numeric
@@ -39,18 +39,21 @@ class Operator(Op):
 
         Op.__init__(self, subject, args)
 
+    def __args__(self):
+        return self.subject, self.args
+
     def __json__(self):
         return to_json(self.forward())
 
-    def __ns__(self, context, name_hint):
-        deanonymize(self.subject, context, name_hint + "_subject")
-        deanonymize(self.args, context, name_hint + "_args")
+    def __ns__(self, cxt, name_hint):
+        cxt.deanonymize(self.subject, name_hint + "_subject")
+        cxt.deanonymize(self.args, name_hint + "_args")
 
-        if is_op_ref(self.subject):
-            self.subject = reference(context, self.subject, name_hint + "_subject")
+        if is_literal(self.subject) or is_op_ref(self.subject):
+            cxt.assign(self.subject, name_hint + "_subject")
 
         if is_op_ref(self.args):
-            self.args = reference(context, self.args, name_hint + "_args")
+            cxt.assign(self.args, name_hint + "_args")
 
     def __repr__(self):
         raise NotImplementedError(f"human-readable string representation of {self.__class__.__name__}")
@@ -65,6 +68,8 @@ class Operator(Op):
 
     @property
     def shape(self):
+        """The shape of the result of this operation"""
+
         raise NotImplementedError(f"{self.__class__.__name__}.shape")
 
     def forward(self):
@@ -110,13 +115,16 @@ class Unary(Operator):
 
         Operator.__init__(self, subject, None)
 
-    def __ns__(self, context, name_hint):
+    def __args__(self):
+        return self.subject,
+
+    def __ns__(self, cxt, name_hint):
         assert self.args is None
 
-        deanonymize(self.subject, context, name_hint + "_subject")
+        cxt.deanonymize(self.subject, name_hint + "_subject")
 
         if is_op_ref(self.subject):
-            self.subject = reference(context, self.subject, name_hint + "_subject")
+            cxt.assign(self.subject, name_hint + "_subject")
 
 
 class Custom(Unary):
@@ -129,9 +137,9 @@ class Custom(Unary):
     def __json__(self):
         return to_json(self._op)
 
-    def __ns__(self, context, name_hint):
-        Unary.__ns__(self, context, name_hint)
-        deanonymize(self._op, context, name_hint + "_custom_op")
+    def __ns__(self, cxt, name_hint):
+        Unary.__ns__(self, cxt, name_hint)
+        cxt.deanonymize(self._op, name_hint + "_custom_op")
 
     @property
     def shape(self):
@@ -172,11 +180,17 @@ class Exp(Unary):
     def forward(self):
         return Numeric.exp(self.subject)
 
-    def backward(self, _variable=None):
-        return self.subject.exp()
+    def backward(self, variable=None):
+        if same_as(variable, self.subject):
+            return derivative_of(self.subject, variable).exp()
+        elif operator(self.subject):
+            # this operator always has exactly one argument, so it's safe to hard-code the single-variable chain rule
+            return derivative_of(self.subject) * self.subject.exp()
+        else:
+            return self.subject.exp()
 
     def gradients(self, loss):
-        return gradients(self.subject, loss * self.backward())
+        return gradients(self.subject, loss * self.subject.exp())
 
     def simplify(self):
         subject = simplify(self.subject)
@@ -408,7 +422,6 @@ class Dual(Operator):
         Operator.__init__(self, subject, args)
 
 
-# TODO: logarithm with an explicit base
 class Log(Operator):
     def __repr__(self):
         return f"log({self.subject})"
@@ -418,14 +431,18 @@ class Log(Operator):
         return self.subject.shape
 
     def forward(self):
-        return Numeric.log(self.subject, self.args)
+        return Numeric.log(self.subject)
 
     def backward(self, variable=None):
-        subject = derivative_of(self.subject, variable) if same_as(self.subject, variable) else self.subject
-        return 1 / subject
+        if variable is None:
+            return chain_rule(self, self.subject, self.args) / self.subject
+        elif same_as(self.subject, variable):
+            return 1 / derivative_of(self.subject, variable)
+        else:
+            return 1 / self.subject
 
     def gradients(self, loss):
-        return gradients(self.subject, loss * self.backward())
+        return gradients(self.subject, loss / self.subject)
 
     def simplify(self):
         subject = simplify(self.subject)
@@ -484,8 +501,10 @@ class Pow(Dual):
     def backward(self, variable=None):
         if same_as(self.args, variable):
             return (self.subject**self.args) * self.subject.log()
-
-        return self.args * (self.subject**(self.args - 1))
+        elif variable is None:
+            return chain_rule(self, self.subject, self.args) * self.args * (self.subject**(self.args - 1))
+        else:
+            return self.args * (self.subject ** (self.args - 1))
 
     def gradients(self, loss):
         subject_grad = loss * self.args * self.subject**(self.args - 1)
@@ -659,30 +678,26 @@ class Div(DualBroadcast):
             return Div(subject, args)
 
 
-def chain_rule(numeric):
+def chain_rule(op, *args):
     """
     Compute the chain rule coefficient of the given :class:`Operator`.
 
     This function will return `1` if the given `numeric` is constant, or has only constant inputs.
     """
 
-    if operator(numeric):
-        op = operator(numeric)
+    if operator(op):
+        op = operator(op)
     else:
-        raise ValueError(f"cannot apply the chain rule to a constant {numeric}")
+        raise ValueError(f"cannot apply the chain rule to a constant {op}")
 
-    # only apply the chain rule for ops that have been explicitly tested
-    if not isinstance(op, (Exp, Log, Pow)):
+    args = [arg for arg in args if operator(arg)]
+
+    if not args:
         return 1
-
-    if operator(op.subject) and operator(op.args):
-        raise NotImplementedError("multivariate chain rule")
-    elif operator(op.subject):
-        return derivative_of(op.subject)
-    elif operator(op.args):
-        return derivative_of(op.args)
+    elif len(args) == 1:
+        return derivative_of(args[0])
     else:
-        return 1
+        return sum(derivative_of(op, arg) * derivative_of(arg) for arg in args)
 
 
 def constant(numeric):
@@ -702,22 +717,33 @@ def constant(numeric):
     return numeric
 
 
-def derivative_of(state, variable=None, keepdims=False):
+def derivative_of(state_or_function, variable=None, keepdims=False):
     """
-    Find the derivative of the given `state`.
+    Find the derivative of the given `state_or_function`.
 
-    If a `variable` is specified, this will be the partial derivative with respect to `variable`.
-    If the given `state` is not differentiable, this will raise a `TypeError`.
+    If a differentiable state is given, this will construct a new op to calculate it derivative, which can be
+    a partial derivative if a `variable` is specified.
+
+    If a differentiable function is given, a new callable function will be returned which computes its derivative.
+
+    If `state_or_function` is not differentiable, a `TypeError` will be raised.
     """
 
-    if not is_numeric(state):
-        raise ValueError(f"cannot take the derivative of a non-numeric state {state} (note the type {type(state)})")
+    if callable(state_or_function):
+        function = state_or_function
+        if hasattr(function, "derivative"):
+            return function.derivative(variable)
+        else:
+            raise ValueError(f"not a differentiable function: {function}")
+
+    state = state_or_function
 
     if same_as(state, variable):
         # it's a partial derivative and this is the free variable
         return ones_like(state, keepdims)
 
-    from ..ml.optimizer import Variable
+    # TODO: it doesn't make sense to import from the ML package in the math package
+    from ..ml.variable import Variable
 
     if isinstance(state, Variable):
         if variable is None:
@@ -730,18 +756,15 @@ def derivative_of(state, variable=None, keepdims=False):
     if is_constant(state):
         return zeros_like(state, keepdims)
     elif operator(state):
-        from ..collection.tensor import Dense, NDArray
-
-        coeff = chain_rule(state)
         d = operator(state).backward(variable)
 
-        if isinstance(state, NDArray) and keepdims:
+        if keepdims:
             if same_as(d, 0):
-                d = zeros_like(state)
+                return zeros_like(state)
             elif same_as(d, 1):
-                d = ones_like(state)
+                return ones_like(state)
 
-        return coeff * d
+        return d
     else:
         raise ValueError(f"the derivative of {state} is not defined")
 
@@ -752,13 +775,15 @@ def gradients(numeric, loss, variables=None):
 
     If one variable is given, one gradient will be returned, or a `KeyError` will be raised if not present in the graph.
     If a list of variables is given, a corresponding list of gradients will be returned.
-    If no variables are given, a :class:`Gradients` object whose keys are the `hex_id` of each input.
+    If no variables are given, a :class:`Gradients` object whose keys are the inputs of the graph will be returned.
     """
 
-    if operator(numeric):
+    if is_literal(numeric):
+        grads = Gradients()
+    elif operator(numeric):
         grads = operator(numeric).gradients(loss)
     elif is_constant(numeric):
-        grads = Gradients({hex_id(numeric): loss})
+        grads = Gradients({numeric: loss})
     elif is_numeric(numeric):
         raise ValueError(f"cannot compute gradients of {numeric} w/r/t {loss}")
     else:
@@ -768,16 +793,16 @@ def gradients(numeric, loss, variables=None):
         return grads
 
     if not isinstance(variables, (list, tuple)):
-        if hex_id(variables) not in grads:
+        if variables not in grads:
             raise KeyError(f"{variables} is not reachable from operator {numeric}")
 
-        return grads[hex_id(variables)]
+        return grads[variables]
 
-    missing = [var for var in variables if hex_id(var) not in grads]
+    missing = [var for var in variables if var not in grads]
     if missing:
         raise KeyError(f"not reachable by traversing the operator graph {numeric}: {missing}")
 
-    return [grads[hex_id(var)] for var in variables]
+    return [grads[var] for var in variables]
 
 
 def is_constant(numeric):
@@ -785,9 +810,6 @@ def is_constant(numeric):
     Return `False` if the given `numeric` state is the result of an :class:`Operator`, i.e. a differentiable function.
     """
 
-    if not is_numeric(numeric):
-        raise TypeError(f"a non-numeric state {numeric} (type {type(numeric)}) cannot be a numeric constant")
-
     return operator(numeric) is None
 
 
@@ -877,32 +899,28 @@ def shape_of(numeric):
 
 def ones_like(state, keepdims=True):
     from ..collection.tensor import Dense
+    from ..scalar.number import Number
 
-    if not keepdims:
-        return 1
-
-    if is_literal(state) or same_as(state.shape.ndim(), 0):
-        if isinstance(state, Numeric):
-            return type(state)(form=1)
-        else:
-            return 1
+    if isinstance(state, Number):
+        return type(state)(form=1)
+    elif is_literal(state) or not keepdims:
+        return Number(1)
     else:
-        return Dense.ones_like(state)
+        cls = type(state) if hasattr(type(state), "ones_like") else Dense
+        return cls.ones_like(state)
 
 
 def zeros_like(state, keepdims=True):
     from ..collection.tensor import Sparse
+    from ..scalar.number import Number
 
-    if not keepdims:
-        return 0
-
-    if is_literal(state) or same_as(state.shape.ndim(), 0):
-        if isinstance(state, Numeric):
-            return type(state)(form=0)
-        else:
-            return 0
+    if isinstance(state, Number):
+        return type(state)(form=0)
+    elif is_literal(state) or not keepdims:
+        return Number(0)
     else:
-        return Sparse.zeros_like(state)
+        cls = type(state) if hasattr(type(state), "zeros_like") else Sparse
+        return cls.zeros_like(state)
 
 
 def debug_shape(numeric):
