@@ -1,5 +1,6 @@
 """Machine learning utilities such as a standard definition of a :class:`NeuralNet` and :class:`Optimizer`"""
 
-from .constants import LIB_URI, sigmoid
+from .activation import sigmoid, softmax, relu
+from .constants import LIB_URI
 from .service import ML
 from .variable import Variable
