@@ -1,7 +1,7 @@
 from ..app import Library
 
 from .constants import LIB_URI
-from .nn import ConvLayer, DNN, DNNLayer, Layer, NeuralNet, Sequential
+from .nn import ConvLayer, DNN, Layer, Linear, NeuralNet, Sequential
 from .optimizer import Adam, GradientDescent, Optimizer
 
 
@@ -16,7 +16,7 @@ class ML(Library):
     # dynamic models (implementations)
     Adam = Adam
     ConvLayer = ConvLayer
-    DNNLayer = DNNLayer
     DNN = DNN
     GradientDescent = GradientDescent
+    Linear = Linear
     Sequential = Sequential
