@@ -1,19 +1,24 @@
+import dataclasses
 import inspect
 
+from ..json import to_json
 from ..scalar import Value
 from ..scalar import op, ref
 from ..state import State
-from ..uri import uri, URI
-from ..context import to_json, Context
+from ..uri import URI
 
-from . import _get_rtype, parse_args, resolve_class
+from .functions import get_rtype, parse_args, resolve_class
 
 
 EMPTY = inspect.Parameter.empty
 
 
 class Method(object):
-    __uri__ = uri(op.Op)
+    __uri__ = URI(op.Op)
+
+    @classmethod
+    def expand(cls, header, form, name):
+        yield name, cls(header, form, name)
 
     def __init__(self, header, form, name):
         if not inspect.isfunction(form):
@@ -22,54 +27,44 @@ class Method(object):
         self.header = header
         self.form = form
         self.name = name
+        self.sig = inspect.signature(self.form)
 
-    def __eq__(self, other):
-        if isinstance(other, Method):
-            return self.form == other.form
-
-        return False
+        if tuple(self.sig.parameters)[0] != "self":
+            raise TypeError(f"method signature must begin with 'self' (found {tuple(self.sig.parameters.items())})")
 
     def __json__(self):
-        return {str(uri(self)): to_json(ref.form_of(self))}
+        return {str(self.__uri__): to_json(ref.form_of(self))}
 
-    def dtype(self):
-        return self.__class__.__name__
+    def __ref__(self):
+        raise NotImplementedError
 
     def subject(self):
-        if isinstance(self.header, State):
-            return ref.MethodSubject(self.header, self.name)
-        else:
-            return uri(self.header).append(self.name)
+        return self.header.__uri__.append(self.name)
 
 
 class Get(Method):
-    __uri__ = uri(op.Get)
+    __uri__ = URI(op.Get)
 
     def __init__(self, header, form, name):
         Method.__init__(self, header, form, name)
-        self.rtype = _get_rtype(self.form, State)
+        self.rtype = get_rtype(self.form, State)
 
     def __call__(self, key=None):
         return self.rtype(form=ref.Get(self.subject(), key))
 
-    def __args__(self):
-        _, cxt = ref.form_of(self)
-        return [self.subject(), cxt]
-
     def __form__(self):
-        cxt, args = _first_params(self)
+        cxt, args = first_params(self)
 
-        sig = inspect.signature(self.form)
         key_name = "key"
-        if len(sig.parameters) == len(args):
+        if len(self.sig.parameters) == len(args):
             pass
-        elif len(sig.parameters) - len(args) == 1:
-            key_name = list(sig.parameters.keys())[-1]
-            param = sig.parameters[key_name]
+        elif len(self.sig.parameters) - len(args) == 1:
+            key_name = tuple(self.sig.parameters.keys())[-1]
+            param = self.sig.parameters[key_name]
             dtype = resolve_class(self.form, param.annotation, Value)
             args.append(dtype(form=URI(key_name)))
         else:
-            raise ValueError(f"{self.dtype()} takes 0-3 parameters: (self, cxt, key)")
+            raise ValueError(f"{self} should 0-3 parameters: (self, cxt, key)")
 
         if key_name in cxt:
             raise RuntimeError(f"namespace collision: {key_name} in {self.form}")
@@ -78,12 +73,23 @@ class Get(Method):
 
         return key_name, cxt
 
+    def __ref__(self, name):
+        sig = tuple(self.sig.parameters.items())
+        assert 0 < len(sig) <= 3
+
+        if len(sig) == 3 or (len(sig) == 2 and sig[1][0] not in ["cxt", "txn"]):
+            ktype = resolve_class(self.form, sig[-1][1].annotation, Value)
+        else:
+            ktype = Value
+
+        return op.Get[ktype, self.rtype](URI(name))
+
 
 class Put(Method):
-    __uri__ = uri(op.Put)
+    __uri__ = URI(op.Put)
 
     def __init__(self, header, form, name):
-        rtype = _get_rtype(form, None)
+        rtype = get_rtype(form, None)
 
         if not ref.is_none(rtype):
             raise ValueError(f"PUT method must return None, not f{rtype}")
@@ -93,23 +99,17 @@ class Put(Method):
     def __call__(self, key, value):
         return ref.Put(self.subject(), key, value)
 
-    def __args__(self):
-        _, _, cxt = ref.form_of(self)
-        return [self.subject(), cxt]
-
     def __form__(self):
-        cxt, args = _first_params(self)
-
-        sig = inspect.signature(self.form)
+        cxt, args = first_params(self)
 
         key_name = "key"
         value_name = "value"
 
-        if len(sig.parameters) == len(args):
+        if len(self.sig.parameters) == len(args):
             pass
-        elif len(sig.parameters) - len(args) == 1:
-            name = list(sig.parameters.keys())[-1]
-            param = sig.parameters[name]
+        elif len(self.sig.parameters) - len(args) == 1:
+            name = tuple(self.sig.parameters.keys())[-1]
+            param = self.sig.parameters[name]
             if name == key_name:
                 dtype = resolve_class(self.form, param.annotation, Value)
                 args.append(dtype(form=URI(key_name)))
@@ -117,19 +117,20 @@ class Put(Method):
                 dtype = resolve_class(self.form, param.annotation, State)
                 args.append(dtype(form=URI(value_name)))
             else:
-                raise ValueError(f"{self.dtype()} with three parameters requires 'key' or 'value', not '{name}'")
-        elif len(sig.parameters) - len(args) == 2:
-            key_name, value_name = list(sig.parameters.keys())[-2:]
+                raise ValueError(f"a PUT method with three parameters requires an explicit 'key' or 'value', not '{name}'")
+        elif len(self.sig.parameters) - len(args) == 2:
+            key_name, value_name = tuple(self.sig.parameters.keys())[-2:]
 
-            param = sig.parameters[key_name]
+            param = self.sig.parameters[key_name]
             dtype = resolve_class(self.form, param.annotation, Value)
             args.append(dtype(form=URI(key_name)))
 
-            param = sig.parameters[value_name]
+            param = self.sig.parameters[value_name]
             dtype = resolve_class(self.form, param.annotation, State)
             args.append(dtype(form=URI(value_name)))
         else:
-            raise ValueError(f"{self.dtype()} requires 0-4 parameters: (self, cxt, key, value)")
+            raise ValueError("a PUT method requires 0-4 parameters: (self, cxt, key, value), " +
+                             f"not {tuple(self.sig.parameters.items())}")
 
         cxt._return = self.form(*args)
 
@@ -139,22 +140,50 @@ class Put(Method):
 
         return key_name, value_name, cxt
 
+    def __ref__(self, name):
+        sig = tuple(self.sig.parameters.items())
+        assert 0 < len(sig) <= 4
+
+        if len(sig) >= 2 and sig[1][0] in ["cxt", "txn"]:
+            sig = sig[2:]
+        else:
+            sig = sig[1:]
+
+        if len(sig) == 1 and sig[0][0] == "key":
+            ktype = resolve_class(self.form, sig[0][1].annotation, Value)
+            vtype = State
+        elif len(sig) == 1 and sig[0][0] == "value":
+            ktype = Value
+            vtype = resolve_class(self.form, sig[0][1].annotation, State)
+        elif len(sig) == 2:
+            ((_, k), (_, v)) = sig
+            ktype = resolve_class(self.form, k.annotation, Value)
+            vtype = resolve_class(self.form, v.annotation, State)
+        elif not sig:
+            ktype = Value
+            vtype = State
+        else:
+            raise ValueError("a PUT method requires 0-4 parameters: (self, cxt, key, value), " +
+                             f"not {tuple(self.sig.parameters.items())}")
+
+        return op.Put[ktype, vtype](URI(name))
+
 
 class Post(Method):
-    __uri__ = uri(op.Post)
+    __uri__ = URI(op.Post)
 
     def __init__(self, header, form, name):
         Method.__init__(self, header, form, name)
-        self.rtype = _get_rtype(self.form, State)
+        self.rtype = get_rtype(self.form, State)
 
     def __call__(self, *args, **kwargs):
-        sig = list(inspect.signature(self.form).parameters.items())
+        sig = tuple(self.sig.parameters.items())
         rtype = self.rtype
 
         if not sig:
-            raise TypeError(f"POST method signature for {self} is missing the 'self' parameter")
+            raise TypeError(f"the method signature for {self} is missing the 'self' parameter")
         elif sig[0][0] != "self":
-            raise TypeError(f"POST method signature must begin with 'self', not '{sig[0][0]}'")
+            raise TypeError(f"a method signature must begin with 'self', not '{sig[0][0]}'")
 
         if len(sig) > 1 and sig[1][0] in ["cxt", "txn"]:
             sig = sig[2:]
@@ -164,18 +193,11 @@ class Post(Method):
         params = parse_args(sig, *args, **kwargs)
         return rtype(form=ref.Post(self.subject(), params))
 
-    def __args__(self):
-        _, cxt = ref.form_of(self)
-        return [self.subject(), cxt]
-
     def __form__(self):
-        cxt, args = _first_params(self)
+        cxt, args = first_params(self)
 
-        sig = inspect.signature(self.form)
         kwargs = {}
-        for name in list(sig.parameters.keys())[len(args):]:
-            param = sig.parameters[name]
-
+        for name, param in tuple(self.sig.parameters.items())[len(args):]:
             dtype = State
             if param.default is inspect.Parameter.empty:
                 if param.annotation:
@@ -194,12 +216,27 @@ class Post(Method):
 
         return cxt
 
+    def __ref__(self, name):
+        fields = []
+
+        sig = tuple(self.sig.parameters.items())
+        sig = sig[2:] if len(sig) > 1 and sig[1][0] in ["cxt", "txn"] else sig[1:]
+        for param_name, param in sig:
+            dtype = resolve_class(self.subject(), param.annotation, State)
+            if param.default is inspect.Parameter.empty:
+                fields.append((param_name, dtype))
+            else:
+                fields.append((param_name, dtype, dataclasses.field(default=param.default)))
+
+        sig = dataclasses.make_dataclass("Args", fields)
+        return op.Post[sig, self.rtype](URI(name))
+
 
 class Delete(Method):
-    __uri__ = uri(op.Delete)
+    __uri__ = URI(op.Delete)
 
     def __init__(self, header, form, name):
-        rtype = _get_rtype(form, None)
+        rtype = get_rtype(form, None)
 
         if not ref.is_none(rtype):
             raise ValueError(f"DELETE method must return None, not f{rtype}")
@@ -209,25 +246,24 @@ class Delete(Method):
     def __call__(self, key=None):
         return ref.Delete(self.subject(), key)
 
-    def __args__(self):
-        _, cxt = ref.form_of(self)
-        return [self.subject(), cxt]
-
     def __form__(self):
         return Get.__form__(self)
 
+    def __ref__(self, name):
+        sig = tuple(self.sig.parameters.items())
+        assert 0 < len(sig) <= 3
 
-class Operator(Post):
-    def __call__(self, *args, **kwargs):
-        result = self.form(self.header, *args, **kwargs)
-
-        if self.rtype:
-            return self.rtype(form=result)
+        if len(sig) == 3 or (len(sig) == 2 and sig[1][0] not in ["cxt", "txn"]):
+            ktype = resolve_class(self.form, sig[-1][1].annotation, Value)
         else:
-            return result
+            ktype = Value
+
+        return op.Delete[ktype](URI(name))
 
 
 def _check_context_param(parameter):
+    from ..context import Context
+
     _name, param = parameter
     if param.annotation == EMPTY or param.annotation == Context:
         pass
@@ -236,20 +272,22 @@ def _check_context_param(parameter):
             f"a method definition takes a transaction context as its second parameter, not {param.annotation}")
 
 
-def _first_params(method):
+def first_params(method):
+    from ..context import Context
+
     sig = inspect.signature(method.form)
 
     if not sig.parameters:
-        raise ValueError(f"{method.dtype()} has at least one argument: (self, cxt, name1=val1, ...)")
+        raise ValueError(f"a method has at least one argument: (self, cxt, name1=val1, ...)")
 
     args = []
 
-    param_names = list(sig.parameters.keys())
+    param_names = tuple(sig.parameters.keys())
 
     if param_names[0] == "self":
         args.append(method.header)
     else:
-        raise ValueError(f"first argument to {method.dtype()} must be 'self', not {param_names[0]}")
+        raise ValueError(f"the first argument to a method must be 'self', not {param_names[0]}")
 
     cxt = Context()
 
